using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Drawing;
using System.Text;
using System.Windows.Forms;

using Syncfusion.Windows.Forms.Grid;
using Syncfusion.Drawing;

namespace POC
{
    public enum GridExportFormat
    {
        Excel,
        Pdf,
        Word,
        Html
    }

    public partial class FFGrid : GridControl
    {
        //If user resized the columns or not. 
        //Till user resized, visible cols have equal width whose total = GridBounds.Width
        public FFGrid()
        {
            InitializeComponent();
            this.Properties.RowHeaders = false;
            this.Model.CommandStack.Enabled = false;
            this.Model.ReadOnly = true;
            this.Model.Options.SelectCellsMouseButtonsMask = MouseButtons.Left;
            this.Model.Options.DisplayEmptyRows = false;
            this.Model.Options.DisplayEmptyColumns = false;
            this.Model.Options.RefreshCurrentCellBehavior = GridRefreshCurrentCellBehavior.RefreshRow;
            this.Model.Options.ResizeColsBehavior = GridResizeCellsBehavior.ResizeAll |
                GridResizeCellsBehavior.OutlineHeaders | GridResizeCellsBehavior.AllowDragOutside;
            this.BaseStylesMap["Header"].StyleInfo.Enabled = false;

            this.Model.ColsHidden += new GridRowColHiddenEventHandler(OnColsHiddenChanged);
            this.Model.ResetVolatileData();
            this.UseProportionateColResizing = true;

            //Support for sorting
            this.CellModels.Add("SortHeader", new GridSortColumnHeaderCellModel(this.Model));

        }

        IVirtualGridDataProvider _dataProvider = null;
        public void WireDataProvider(IVirtualGridDataProvider dataProvider)
        {
            UnwireDataProvider();

            this.Selections.Clear();
            _dataProvider = dataProvider;
            _dataProvider.ResetVolatileData += new EventHandler(OnResetVolatileData);

        }

        public void UnwireDataProvider()
        {
            if (_dataProvider != null)
            {
                _dataProvider.ResetVolatileData -= new EventHandler(OnResetVolatileData);
                _dataProvider = null;
            }
        }

        private bool _hilightAlternateRows = true;
        public bool HilightAlternateRows
        {
            get { return _hilightAlternateRows; }
            set { _hilightAlternateRows = value; }
        }

        private Color _hilightAlternateRowColor = Color.PowderBlue;
        public Color HilightAlternateRowColor
        {
            get { return _hilightAlternateRowColor; }
            set { _hilightAlternateRowColor = value; }
        }

        private Color _headerBackColor = Color.Teal;
        public Color HeaderBackColor
        {
            get { return _headerBackColor; }
            set { _headerBackColor = value; }
        }

        private Color _headerForeColor = SystemColors.Info;
        public Color HeaderForeColor
        {
            get { return _headerForeColor; }
            set { _headerForeColor = value; }
        }

        private bool _colRatiosAvailable = false;
        private double[] colRatios;
        
        private bool _useProportionateColResizing = false;
        public bool UseProportionateColResizing
        {
            get 
            { 
                return _useProportionateColResizing; 
            }
            set 
            {
                _useProportionateColResizing = value;

                if (value)
                {
                    if (colRatios == null || colRatios.Length != this.Model.ColCount)
                        colRatios = new double[this.Model.ColCount];

                    CalculateColRatios();
                }
                else
                {
                    Refresh();
                }
            }
        }

        private Styler _styler = null;
        [Browsable(false), Bindable(false)]
        public Styler GridStyler
        {
            get
            {
                return _styler;
            }
            set
            {
                if (value != null)
                {
                    _styler = value;
                    _styler.GridData = this.Model;
                    this.Refresh(true);
                }
            }
        }

        private bool _allowSoring = false;
        public bool AllowSorting
        {
            get { return _allowSoring; }
            set 
            {
                _allowSoring = value;
            }
        }

        private int _sortColumn = 1;
        private int SortColumn
        {
            get { return _sortColumn; }
            set 
            { 
                _sortColumn = value;
            }
        }

        private ListSortDirection _defaultSortDirection = ListSortDirection.Ascending;
        public ListSortDirection DefaultSortDirection
        {
            get { return _defaultSortDirection; }
            set 
            { 
                _defaultSortDirection = value;
            }
        }

        private Dictionary<int, bool> _hiddenColumns = new Dictionary<int, bool>();
        public void SetColumHidden(int colIndex, bool donotShowInHideShowColumns)
        {
            this.Model.HideCols[colIndex] = true;
            _hiddenColumns[colIndex] = donotShowInHideShowColumns;
        }

        #region virtual grid overrides

        protected override void OnQueryRowCount(GridRowColCountEventArgs e)
        {
            if (_dataProvider != null)
                _dataProvider.QueryRowCount(e);

            base.OnQueryRowCount(e);
        }

        private int _prevColCount = 0;
        protected override void OnQueryColCount(GridRowColCountEventArgs e)
        {
            if (_dataProvider != null)
                _dataProvider.QueryColCount(e);

            base.OnQueryColCount(e);

            if (e.Count != _prevColCount)
            {
                _prevColCount = e.Count;
                if (_columnInfos != null)
                    _columnInfos.Clear();

                this.UseProportionateColResizing = this.UseProportionateColResizing;
            }
        }

        protected override void OnQueryCellInfo(GridQueryCellInfoEventArgs e)
        {
            if (e.RowIndex == 0)
            {
                e.Style.BackColor = this.HeaderBackColor;
                e.Style.TextColor = this.HeaderForeColor;
            }

            if (e.RowIndex != 0 && _hilightAlternateRows && e.RowIndex % 2 == 0)
                e.Style.BackColor = this.HilightAlternateRowColor;

            e.Style.WrapText = true;
            e.Style.Trimming = StringTrimming.EllipsisCharacter;
            if (_dataProvider != null)
                _dataProvider.QueryCellInfo(e);
            base.OnQueryCellInfo(e);
            e.Style.CellTipText = e.Style.Text;

            if (e.RowIndex == 0)
            {
                this.HeaderBackColor = e.Style.BackColor;
                this.HeaderForeColor = e.Style.TextColor;
            }

            if (e.RowIndex != 0 && _hilightAlternateRows && e.RowIndex % 2 == 0)
                this.HilightAlternateRowColor = e.Style.BackColor;
        }

        protected override void OnQueryRowHeight(GridRowColSizeEventArgs e)
        {
            if (_dataProvider != null)
                _dataProvider.QueryRowHeight(e);
            base.OnQueryRowHeight(e);
        }

        protected override void OnQueryColWidth(GridRowColSizeEventArgs e)
        {
            if (!e.Handled && _dataProvider != null)
                _dataProvider.QueryColWidth(e);
            base.OnQueryColWidth(e);

            if (!e.Handled && this.UseProportionateColResizing)
            {
                if (_colRatiosAvailable)
                {
                    if (e.Index > 0)
                    {
                        e.Size = (int)(colRatios[e.Index - 1] * this.ClientSize.Width);
                        e.Handled = true;
                    }
                }
            }
        }

        protected override void OnQueryCoveredRange(GridQueryCoveredRangeEventArgs e)
        {
            if (_dataProvider != null)
                _dataProvider.QueryCoveredRange(e);
            base.OnQueryCoveredRange(e);
        }

        //protected override void Refresh(bool fromModel)
        //{
        //    this.Model.ResetVolatileData();
        //    base.Refresh(fromModel);
        //}

        #endregion

        protected override void OnKeyDown(KeyEventArgs e)
        {
            if(e.KeyData == Keys.Escape)
            {
                this.Model.Selections.Clear(true);
                e.Handled = true;
                this.Refresh();
            }
            base.OnKeyDown(e);
        }

        protected override void OnClick(EventArgs e)
        {
            Point pt = this.PointToClient(Control.MousePosition);
            int row, col;
            this.PointToRowCol(pt, out row, out col);

            if(this.AllowSorting && row == 0 && col != 0)
            {
                DoSorting(col);
            }
            else if (row > this.Model.RowCount || row == -1 || col == -1)
            {
                //Not clicked on any of the rows, so remove all selections.
                this.Model.Selections.Clear(true);
                this.Refresh();
            }
        }

        protected override void OnPrepareViewStyleInfo(GridPrepareViewStyleInfoEventArgs e)
        {
            base.OnPrepareViewStyleInfo(e);
            if (e.RowIndex == 0)
            {
                e.Style.BackColor = this.HeaderBackColor;
                e.Style.TextColor = this.HeaderForeColor;
                e.Style.CellAppearance = GridCellAppearance.Raised;
            }

            if (_styler != null)
            {
                _styler.SetInfo(e.RowIndex, e.ColIndex, e.Style);
            }
            else
            {
                if (e.RowIndex != 0 && _hilightAlternateRows && e.RowIndex % 2 == 0)
                    e.Style.BackColor = this.HilightAlternateRowColor;
            }

            //To highlight the whole row when the current cell is being displayed.
            if (this.Model.Selections.Count > 0 && e.RowIndex > this.Model.Rows.HeaderCount && e.ColIndex > this.Model.Cols.HeaderCount && this.CurrentCell.HasCurrentCellAt(e.RowIndex))
            {
                e.Style.Interior = new BrushInfo(SystemColors.Highlight);
                e.Style.TextColor = SystemColors.HighlightText;
                //e.Style.Font.Bold = true;
            }
        }

        protected override void OnSelectionChanged(GridSelectionChangedEventArgs e)
        {
            base.OnSelectionChanged(e);
            //No Selection, So deactivate the current row.
            if (e.Range.Height <= 0 && e.Range.Width <= 0)
                RefreshSelections(e.OldRanges.ActiveRange.Bottom);
        }

        protected override void OnCurrentCellDeactivated(GridCurrentCellDeactivatedEventArgs e)
        {
            base.OnCurrentCellDeactivated(e);

            GridCurrentCell cc = this.CurrentCell;
            if (!cc.IsInMoveTo || cc.MoveToRowIndex != cc.MoveFromRowIndex)
                RefreshSelections(e.RowIndex);
        }

        protected override void OnVScrollBarValueChanged(object sender, EventArgs e)
        {
            base.OnVScrollBarValueChanged(sender, e);
            AdjustRowHeights();
        }

        protected override void OnClientSizeChanged(EventArgs e)
        {
            base.OnClientSizeChanged(e);
            AdjustRowHeights();
        }

        protected override void OnResizingRows(GridResizingRowsEventArgs e)
        {
            if (e.Reason == GridResizeCellsReason.DoubleClick)
                e.Cancel = true;
            else if (this.UseProportionateColResizing && e.Reason == GridResizeCellsReason.HitTest)
                e.Cancel = true;
            else
                base.OnResizingRows(e);
        }

        protected override void OnResizingColumns(GridResizingColumnsEventArgs e)
        {
            //Double Click actually shows the hidden cloumns. So cancel it.
            if (e.Reason == GridResizeCellsReason.DoubleClick && !this.UseProportionateColResizing)
            {
                this.ColWidths.ResizeToFit(this.ViewLayout.VisibleCellsRange, GridResizeToFitOptions.IncludeHeaders);
                e.Cancel = true;
            }
            else if (this.UseProportionateColResizing && e.Reason == GridResizeCellsReason.HitTest)
                e.Cancel = true;
            else
                base.OnResizingColumns(e);
        }

        protected virtual void OnResetVolatileData(object sender, EventArgs e)
        {
            if (this.InvokeRequired)
            {
                this.Invoke(Delegate.CreateDelegate(typeof(EventHandler), this, "OnResetVolatileData"), sender, e);
            }
            else
            {
                this.Model.ResetVolatileData();
                CalculateColRatios();
                //this.Refresh();
            }
        }

        
        private void OnColsHiddenChanged(object sender, GridRowColHiddenEventArgs e)
        {
            CalculateColRatios();
        }

        private Dictionary<ColumnInfo, bool> _columnInfos = null;
        private Dictionary<ColumnInfo, bool> Columns
        {
            get
            {
                if (_columnInfos == null || _columnInfos.Count == 0)
                {
                    if (_columnInfos == null)
                        _columnInfos = new Dictionary<ColumnInfo, bool>(this.Model.ColCount);
                    for (int i = 1; i <= this.Model.ColCount; i++)
                    {
                        if (_hiddenColumns.ContainsKey(i) && _hiddenColumns[i])
                            continue;

                        _columnInfos[new ColumnInfo(this.Model[0, i].CellValue.ToString(), i)] = !this.Model.HideCols[i];

                    }
                }

                return _columnInfos;
            }
        }

        private void DoSorting(int newSortColumn)
        {
            if (newSortColumn > this.Model.ColCount)
                return;

            ListSortDirection sortDir = GetSortDirection(newSortColumn);
            if(newSortColumn != this.SortColumn)
            {
                this.Model[0, this.SortColumn].CellType = "Header";
                this.Model[0, newSortColumn].CellType = "SortHeader";
                this.SortColumn = newSortColumn;
            }
            else
            {
                sortDir = sortDir == ListSortDirection.Ascending ? ListSortDirection.Descending : ListSortDirection.Ascending;
            }

            this.Model[0, SortColumn].Tag = sortDir;
            this.Model.Data.SortByColumn(SortColumn, sortDir);
            this.Refresh();
        }

        private ListSortDirection GetSortDirection(int col)
        {
            if (this.Model[0, col].HasTag)
               return (ListSortDirection)this.Model[0, col].Tag;

           return this.DefaultSortDirection;
        }

        private void CalculateColRatios()
        {
            if (this.UseProportionateColResizing && this.Model.ColCount > 0)
            {
                _colRatiosAvailable = false;
                this.Model.BeginUpdate();
                this.Model.ColWidths.ResizeToFit(this.ViewLayout.VisibleCellsRange, GridResizeToFitOptions.IncludeHeaders);
                double gridwidth = this.Model.Cols.Size.GetTotal(1, this.Model.ColCount);

                for (int i = 1; i <= this.Model.ColCount; i++)
                    colRatios[i - 1] = (double)(this.Model.ColWidths[i] / gridwidth);
                this.Model.EndUpdate(true);
                _colRatiosAvailable = true;
                AdjustRowHeights();
            }
        }

        private void AdjustRowHeights()
        {
            if (this.UseProportionateColResizing)
            {
                this.Model.BeginUpdate();
                this.Model.RowHeights.ResizeToFit(GridRangeInfo.Row(0));
                this.Model.RowHeights.ResizeToFit(this.ViewLayout.VisibleCellsRange);
                this.Model.EndUpdate(true);
                this.Model.ResetVolatileData();
                base.Refresh();
            }
        }

        private void RefreshSelections(int rowIndex)
        {
            this.RefreshRange(GridRangeInfo.Row(rowIndex), GridRangeOptions.MergeAllSpannedCells);
        }
    }

    internal class ColumnInfo
    {
        internal ColumnInfo(string name, int col)
        {
            Name = name;
            Column = col;
        }

        internal string Name;
        internal int Column;

        public override string ToString()
        {
            return Name;
        }
    }
}

//public FFGrid()
//{
//    //Tri state sort support - ascending, Descending, None
//    //this.Model.BaseStylesMap["Column Header"].StyleInfo.CellType = "Header";
//    //this.Model.Data.RowCount = 1;
//    //this.Model.Rows.HeaderCount = 1;
//    //this.Model.Rows.FrozenCount = 1;
//    //this.TriStateSorting = false;
//    //this.Binder.DataSourceChanged += new EventHandler(OnDataSourceChanged);
//    //this.Binder.BindingListChanged += new ListChangedEventHandler(OnGridDataChanged);
//}

//public bool TriStateSorting
//{
//    get
//    {
//        return this.Model.HideRows[0];
//    }
//    set
//    {
//        EnableTriStateSorting(value);
//    }
//}

//GridFilterBar _filterBar = null;
//public bool ShowFilterBar
//{
//    get
//    {
//        return _filterBar != null && _filterBar.Wired ? true : false;
//    }
//    set
//    {
//        if (value)
//        {
//            if (_filterBar == null)
//                _filterBar = new GridFilterBar();
//            if (!_filterBar.Wired)
//                _filterBar.WireGrid(this);
//        }
//        else
//        {
//            if (_filterBar == null || !_filterBar.Wired)
//                return;
//            _filterBar.UnwireGrid();
//        }
//    }
//}

//public GridStyleInfo FilterBarStyleInfo
//{
//    get
//    {
//        if (_filterBar == null)
//            _filterBar = new GridFilterBar();
//        return _filterBar.StyleInfo;
//    }
//    set
//    {
//        if (_filterBar == null)
//            _filterBar = new GridFilterBar();
//        _filterBar.StyleInfo = value;
//    }
//}

//private void OnGridDataChanged(object sender, ListChangedEventArgs e)
//{
//    this.Model.ColWidths.ResizeToFit(GridRangeInfo.Cols(0, this.Model.ColCount), GridResizeToFitOptions.IncludeHeaders);
//}

//private void OnDataSourceChanged(object sender, EventArgs e)
//{
//    this.Model.ColWidths.ResizeToFit(GridRangeInfo.Cols(0, this.Model.ColCount), GridResizeToFitOptions.IncludeHeaders);

//    //if (!TriStateSorting)
//    //    return;

//    //for (int col = 1; col <= this.Model.ColCount; ++col)
//    //{
//    //    this[1, col] = this[0, 1];
//    //    this[1, col].CellType = "ColumnHeaderCell";
//    //}
//}

//private void EnableTriStateSorting(bool enable)
//{
//    if (enable)
//    {
//        this.Model.HideRows[0] = true;
//        this.Model.HideRows[1] = false;
//    }
//    else
//    {
//        this.Model.HideRows[0] = false;
//        this.Model.HideRows[1] = true;
//    }
//}

//private void OnCellClick(object sender, Syncfusion.Windows.Forms.Grid.GridCellClickEventArgs e)
//{
//    //if (this[e.RowIndex, e.ColIndex].CellType == "ColumnHeaderCell")
//    //{
//    //    if (this[e.RowIndex, e.ColIndex].HasTag)
//    //    {
//    //        if (this[e.RowIndex, e.ColIndex].Tag.ToString() == "None")
//    //        {
//    //            this[e.RowIndex, e.ColIndex].Tag = ListSortDirection.Ascending;
//    //            base.SortColumn(e.ColIndex);
//    //        }
//    //        else if ((ListSortDirection)this[e.RowIndex, e.ColIndex].Tag == ListSortDirection.Ascending)
//    //        {
//    //            this[e.RowIndex, e.ColIndex].Tag = ListSortDirection.Descending;
//    //            base.SortColumn(e.ColIndex);
//    //        }
//    //        else if ((ListSortDirection)this[e.RowIndex, e.ColIndex].Tag == ListSortDirection.Descending)
//    //        {
//    //            this[e.RowIndex, e.ColIndex].Tag = "None";
//    //            base.SortColumn(e.ColIndex);
//    //        }
//    //    }
//    //    else
//    //    {
//    //        this[e.RowIndex, e.ColIndex].Tag = ListSortDirection.Ascending;
//    //        base.SortColumn(e.ColIndex);
//    //    }
//    //}
//}